#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <resolv.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "lua.h"
#include "tcp.h"


int tcp_open(const char *host, const char *port, int type)
{
  int sd; /* socket descriptor */
  int port_number; /* port number for socket */
  char *service_name; /* name of service for port */

  /* get port number */
  port_number = resolve_service(port, "tcp", &service_name);

  if(port < 0)
    return TCP_UNKNOWNSERV;

  /* check connection type */
  if(type == 1)
    sd = create_server_socket(host, port_number);
  else
    sd = create_client_socket(host, port_number);
  
  return sd;
}

int resolve_service(const char *port, char *protocol, char **service_name)
{
  struct servent *servent;
  int port_number;

  if(is_number(port))
  {
    port_number = atoi(port);

    if(service_name != NULL)
    {
      servent = getservbyport(htons(port_number), "tcp");
      if(servent != NULL)
        *service_name = servent->s_name;
      else
        *service_name = NULL;
    }
    return port_number;
  }
  else /* get port number from service name */
  {
    servent = getservbyname(port, "tcp");
    
    if(servent == NULL)
      return -1;
    
    if(service_name != NULL)
      *service_name = servent->s_name;

    return ntohs(servent->s_port);
  }
}

int is_number(const char *s)
{
  while(*s)
  {
    if(*s < '0' || *s > '9')
      return 0;

    s++;
  }
  return 1;
}

int create_server_socket(const char *host, int port)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  long addr;
  int sd;
  int on;
  struct linger linger;
  
  bzero(&sa, sizeof(sa));

  if((addr = inet_addr(host)) != -1 )
  {
    /* is internet address in octet notation */
    bcopy(&addr, (char *)&sa.sin_addr, sizeof(addr)); /* set address */
    sa.sin_family = AF_INET;
  }
  else
  {
    if((hp = gethostbyname(host)) == NULL)
      return TCP_UNKNOWNHOST;

    bcopy(hp->h_addr, (char *)&sa.sin_addr, hp->h_length);
    sa.sin_family = hp->h_addrtype;
  }

  sa.sin_port = htons((u_short)port);

  /* open socket */
  if((sd = socket(sa.sin_family, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return TCP_ERROR;

  /* turn off bind address checking */
  on = 1;
  if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on)) == -1)
    return TCP_ERROR;

  /* make socket linger */
  linger.l_onoff = 1;
  linger.l_linger = 30;
  if(setsockopt(sd, SOL_SOCKET, SO_LINGER, (const char *)&linger, sizeof(linger)) == -1)
    return TCP_ERROR;

  /* bind socket to address */
  if(bind(sd, (struct sockaddr *)&sa, sizeof(sa)) == -1)
    return TCP_ERROR;

  /* make socket listen for connections */
  if(listen(sd, MAX_QUEUED_CONNECTIONS) == -1)
    return TCP_ERROR;

  return sd;
}

int create_client_socket(const char *host, int port)
{
  struct sockaddr_in sa;
  struct hostent *hp;
  long addr;
  int sd;
  int on;
  struct linger linger;

  bzero(&sa, sizeof(sa));

  if((addr = inet_addr(host)) != -1)
  {
    /* is internet addr in octet notation */
    bcopy(&addr, (char *)&sa.sin_addr, sizeof(addr)); // set address
    sa.sin_family = AF_INET;
  }
  else
  {
    /* do we know the host's address? */
    if((hp = gethostbyname(host)) == NULL)
      return TCP_UNKNOWNHOST;
    
    bcopy(hp->h_addr, (char *)&sa.sin_addr, hp->h_length);
    sa.sin_family = hp->h_addrtype;
  }

  sa.sin_port = htons((u_short)port);

  /* open socket */
  if((sd = socket(sa.sin_family, SOCK_STREAM, IPPROTO_TCP)) < 0)
    return TCP_ERROR;

  /* turn off bind address checking */
  on = 1;
  if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on)) == -1)
    return TCP_ERROR;

  /* make socket linger */
  linger.l_onoff = 1;
  linger.l_linger = 30;
  if(setsockopt(sd, SOL_SOCKET, SO_LINGER, (const char *)&linger, sizeof(linger)) == -1)
    return TCP_ERROR;

  /* connect to host */
  if(connect(sd, (struct sockaddr *)&sa, sizeof(sa)) < 0)
  {
    close(sd);
    return TCP_ERROR;
  }

  return sd;
}

/* checks if a socket has data waiting */
int tcp_check(int sd, int sleeptime)
{
  fd_set read_fds;
  int max_fd;
  struct timeval delay;

  /* set up our fd set */
  FD_ZERO(&read_fds);
  FD_SET(sd, &read_fds);
  max_fd = sd + 1;

  /* set the delay */
  delay.tv_sec = 0;
  delay.tv_usec = sleeptime; /* 1,000,000 usecs in one second */

  /* take a nap */
  if(select(max_fd, &read_fds, NULL, NULL, &delay) < 0)
    return TCP_ERROR; /* select failed, just quietly bail */

  /* anything there? */
  if(FD_ISSET(sd, &read_fds))
    return TCP_OK; /* yep, return ok */
  else
    return TCP_ERROR; /* nope, return nothing */
}

/* accept a connection on a socket */
int tcp_accept(int sd, char **host)
{
  struct sockaddr_in a_name;
  unsigned int sa_size = sizeof(a_name);
  int accept_sd;

  memset(&a_name, 0, sa_size);

  if((accept_sd = accept(sd, (struct sockaddr *)&a_name, &sa_size)) > 0)
  {
    if(getpeername(accept_sd, (struct sockaddr *)&a_name, &sa_size) == -1)
      *host = NULL;
    else
      *host = (char *)strdup((char *)inet_ntoa(a_name.sin_addr));
  }

  return accept_sd;
}

int tcp_send(int sd, const char *data, int bytes)
{
  int bytes_sent;

  bytes_sent = write(sd, data, bytes);

  if(bytes_sent == -1 || bytes_sent != bytes)
    return TCP_ERROR;

  return bytes_sent;
}

int tcp_recv(int sd, char **data)
{
  int bytes_read;
  char buffer[BUFFER_SIZE];

  bytes_read = read(sd, buffer, BUFFER_SIZE);

  if(bytes_read == -1)
    return TCP_ERROR;
  else if(bytes_read == 0)
    return TCP_EOF;
  else
  {
    *data = buffer;
    return bytes_read;
  }
}

int tcp_close(int sd)
{
  if(sd <= 0)
    return TCP_BADSD;

  close(sd);
  return TCP_OK;
}
